/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1999
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Seth Spitzer <sspitzer@netscape.com>
 *   Robert Ginda <rginda@netscape.com>
 *   Justin Arthur <justinarthur@ieee.org>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*
 * This file contains the following MozPhone related components:
 * 1. Command line handler service, for responding to the -phone command line
 *    option.
 * 2. Protocol handler for supplying a channel to the browser when an tel://
 *
 * Borrowed from ChatZilla. Thanks !
 */

function debug(s) { dump(s); }

/* components defined in this file
	MozPhone UUID is 4de2cc0c-a9e4-4853-871c-7cda72582e13 */
const CLINE_SERVICE_CONTRACTID =
    "@mozilla.org/commandlinehandler/general-startup;1?type=phone";
const CLINE_SERVICE_CID =
    Components.ID("{da2a5256-8e9c-4b05-a9cc-d80e249fd493}");
const TELCONTENT_LISTENER_CONTRACTID =
    "@mozilla.org/uriloader/tel-external-content-listener;1";
const TELCONTENT_LISTENER_CID =
    Components.ID("{ca83f67b-d850-48d1-afa2-057b6d3969f0}");
const TELPROT_HANDLER_CONTRACTID =
    "@mozilla.org/network/protocol;1?name=tel";
const TELPROT_HANDLER_CID =
    Components.ID("{bb225d9f-4c7d-40a8-aa0a-c758c875b5e2}");

const TEL_MIMETYPE = "application/x-tel";

/* components used in this file */
const MEDIATOR_CONTRACTID =
    "@mozilla.org/appshell/window-mediator;1";
const STANDARDURL_CONTRACTID = 
    "@mozilla.org/network/standard-url;1";
const IOSERVICE_CONTRACTID = 
    "@mozilla.org/network/io-service;1";
const ASS_CONTRACTID =
    "@mozilla.org/appshell/appShellService;1";
const RDFS_CONTRACTID =
    "@mozilla.org/rdf/rdf-service;1";

/* interafces used in this file */
const nsIWindowMediator  = Components.interfaces.nsIWindowMediator;
const nsICmdLineHandler  = Components.interfaces.nsICmdLineHandler;
const nsICategoryManager = Components.interfaces.nsICategoryManager;
const nsIContentHandler  = Components.interfaces.nsIContentHandler;
const nsIURIContentListener = Components.interfaces.nsIURIContentListener;
const nsIURILoader       = Components.interfaces.nsIURILoader;
const nsIProtocolHandler = Components.interfaces.nsIProtocolHandler;
const nsIURI             = Components.interfaces.nsIURI;
const nsIStandardURL     = Components.interfaces.nsIStandardURL;
const nsIChannel         = Components.interfaces.nsIChannel;
const nsIRequest         = Components.interfaces.nsIRequest;
const nsIIOService       = Components.interfaces.nsIIOService;
const nsIAppShellService = Components.interfaces.nsIAppShellService;
const nsISupports        = Components.interfaces.nsISupports;
const nsISupportsWeakReference = Components.interfaces.nsISupportsWeakReference;
const nsIRDFService      = Components.interfaces.nsIRDFService;

/* Command Line handler service */
function CLineService()
{}

CLineService.prototype.commandLineArgument = "-phone";
CLineService.prototype.prefNameForStartup = "general.startup.phone";
CLineService.prototype.chromeUrlForTask = "chrome://mozphone/content";
CLineService.prototype.helpText = "Start with an IAX client";
CLineService.prototype.handlesArgs = false;
CLineService.prototype.defaultArgs = "";
CLineService.prototype.openWindowWithArgs = false;

/* factory for command line handler service (CLineService) */
var CLineFactory = new Object();

CLineFactory.createInstance =
function (outer, iid)
{
debug('clinefactory_createInstance\n');
    if (outer != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;

    if (!iid.equals(nsICmdLineHandler) && !iid.equals(nsISupports))
        throw Components.results.NS_ERROR_INVALID_ARG;

    return new CLineService();
}

/* content listener */
function TELContentListener(isSecure)
{
    this.isSecure = isSecure;
}

TELContentListener.prototype.QueryInterface =
function telcl_QueryInterface(iid)
{
debug('telcl_QueryInterface\n');
    if (!iid.equals(nsIURIContentListener) && 
        !iid.equals(nsISupportsWeakReference) &&
        !iid.equals(nsISupports))
    {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }

    return this;
}

TELContentListener.prototype.loadCookie = null;
TELContentListener.prototype.parentContentListener = null;

TELContentListener.prototype.onStartURIOpen =
function telcl_onStartURIOpen(uri)
{
}

TELContentListener.prototype.doContent =
function telcl_doContent(contentType, preferred, request, contentHandler, count)
{
debug( 'telcl_doContent\n' );
    var e;
    var channel = request.QueryInterface(nsIChannel);

    var wmClass = Components.classes[MEDIATOR_CONTRACTID];
    var windowManager = wmClass.getService(nsIWindowMediator);
    
    var assClass = Components.classes[ASS_CONTRACTID];
    var ass = assClass.getService(nsIAppShellService);
    hiddenWin = ass.hiddenDOMWindow;
    
    // Ok, not starting currently, so check if we've got existing windows.
    var w = windowManager.getMostRecentWindow("tel:mozphone");
    
    // Claiming that a MozPhone window is loading.
    if ("MozPhoneStarting" in hiddenWin) {
        dump("mp-service: Mozphone claiming to be starting.\n");
debug("mp-service: Mozphone claiming to be starting.\n");
        if (w && ("client" in w) && ("initialized" in w.client) && 
            w.client.initialized)
        {
            dump("mp-service: It lied. It's finished starting.\n");
debug("mp-service: It lied. It's finished starting.\n");
            // It's actually loaded ok.
            delete hiddenWin.MozPhoneStarting;
        }
    }
    
    if ("MozPhoneStarting" in hiddenWin) {
        count = count || 0;
        
        if ((new Date() - hiddenWin.MozPhoneStarting) > 10000) {
            dump("mp-service: Continuing to be unable to talk to existing window!\n");
        } else {
            //dump("cz-service: **** Try: " + count + ", delay: " + (new Date() - hiddenWin.ChatZillaStarting) + "\n");
            
            // We have a MozPhone window, but we're still loading.
            hiddenWin.setTimeout(function wrapper(t, count) {
                    t.doContent(contentType, preferred, request, contentHandler, count + 1);
                }, 250, this, count);
            return true;
        }
    }

    // MozPhone lives !
    if (w) {
        dump("mp-service: Existing, fully loaded window. Using.\n");
debug("mp-service: Existing, fully loaded window. Using.\n");
        // Mozphone is working and initialized ok. Use it.
        var num = channel.URI.host;
        w.focus();
        w.gui.call_adr.value = num;
        w.call();
        return true;
    }
    dump("mp-service: No windows, starting new one.\n");
    // Ok, no available window, loading or otherwise, so start MozPhone.
    var args = new Object();
    args.url = channel.URI.host;
 
    hiddenWin.MozPhoneStarting = new Date();
    hiddenWin.openDialog("chrome://mozphone/content/mozphone.xul",
       "_blank", "chrome,menubar,toolbar,status,resizable,dialog=no", args);
    return true;
}

TELContentListener.prototype.isPreferred =
function telcl_isPreferred(contentType, desiredContentType)
{
debug('telcl_isPreferred\n');
    return (contentType == TEL_MIMETYPE );
}

TELContentListener.prototype.canHandleContent =
function telcl_canHandleContent(contentType, isContentPreferred, desiredContentType)
{
debug('telcl_canHandleContent\n');
    return (contentType == TEL_MIMETYPE );
}

/* protocol handler factory object (TELContentListener) */
var TELContentListenerFactory = new Object();

TELContentListenerFactory.createInstance =
function telclf_createInstance(outer, iid)
{
debug('telclf_createInstance\n');
    if (outer != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;
    
    if (!iid.equals(nsIURIContentListener) && 
        !iid.equals(nsISupportsWeakReference) &&
        !iid.equals(nsISupports))
    {
        throw Components.results.NS_ERROR_INVALID_ARG;
    }
    
    return new TELContentListener(false);
}


/* tel protocol handler component */
function TELProtocolHandler(isSecure)
{
debug('TELProtocolHandler\n');
    this.isSecure = isSecure;
}

TELProtocolHandler.prototype.protocolFlags = 
                   nsIProtocolHandler.URI_NORELATIVE;

TELProtocolHandler.prototype.allowPort =
function telph_allowPort(port, scheme)
{
debug('telph_allowPort\n');
    return false;
}

TELProtocolHandler.prototype.newURI =
function telph_newURI(spec, charset, baseURI)
{
debug('telph_newURI\n');
    var cls = Components.classes[STANDARDURL_CONTRACTID];
    var url = cls.createInstance(nsIStandardURL);
    url.init(nsIStandardURL.URLTYPE_STANDARD, (this.isSecure ? 9999 : 6667), spec, charset, baseURI);

    return url.QueryInterface(nsIURI);
}

TELProtocolHandler.prototype.newChannel =
function telph_newChannel(URI)
{
debug('telph_newChannel\n');
    ios = Components.classes[IOSERVICE_CONTRACTID].getService(nsIIOService);
    if (!ios.allowPort(URI.port, URI.scheme))
        throw Components.results.NS_ERROR_FAILURE;
    
    var bogusChan = new BogusChannel(URI, this.isSecure);
    bogusChan.contentType = TEL_MIMETYPE;
    return bogusChan;
}

/* protocol handler factory object (TELProtocolHandler) */
var TELProtocolHandlerFactory = new Object();

TELProtocolHandlerFactory.createInstance =
function telphf_createInstance(outer, iid)
{
debug('telphf_createInstance\n');
    if (outer != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;
    
    if (!iid.equals(nsIProtocolHandler) && !iid.equals(nsISupports))
        throw Components.results.NS_ERROR_INVALID_ARG;
    
    var protHandler = new TELProtocolHandler(false);
    protHandler.scheme = "tel";
    protHandler.defaultPort = -1;
    return protHandler;
}

/* bogus IRC channel used by the TELProtocolHandler */
function BogusChannel(URI, isSecure)
{
debug('BogusChannel\n');
    this.URI = URI;
    this.originalURI = URI;
    this.isSecure = isSecure;
}

BogusChannel.prototype.QueryInterface =
function bc_QueryInterface(iid)
{
debug('bc_QueryInterface\n');
if (!iid.equals(nsIChannel) && !iid.equals(nsIRequest) &&
    !iid.equals(nsISupports))
    debug('Components.results.NS_ERROR_NO_INTERFACE\n');
    
    if (!iid.equals(nsIChannel) && !iid.equals(nsIRequest) &&
        !iid.equals(nsISupports))
        throw Components.results.NS_ERROR_NO_INTERFACE;
    
    return this;
}

/* nsIChannel */
BogusChannel.prototype.loadAttributes = null;
BogusChannel.prototype.contentLength = 0;
BogusChannel.prototype.owner = null;
BogusChannel.prototype.loadGroup = null;
BogusChannel.prototype.notificationCallbacks = null;
BogusChannel.prototype.securityInfo = null;

BogusChannel.prototype.open =
function bc_open()
{
debug('bc_open\n');
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
}

BogusChannel.prototype.asyncOpen =
function bc_asyncOpen(observer, ctxt)
{
debug('bc_asyncOpen\n');
    observer.onStartRequest(this, ctxt);
}

BogusChannel.prototype.asyncRead =
function bc_asyncRead(listener, ctxt)
{
debug('bc_asyncRead\n');
    return listener.onStartRequest(this, ctxt);
}

/* nsIRequest */
BogusChannel.prototype.isPending =
function bc_isPending()
{
debug('bc_isPending\n');
    return true;
}

BogusChannel.prototype.status = Components.results.NS_OK;

BogusChannel.prototype.cancel =
function bc_cancel(status)
{
debug('bc_cancel\n');
    this.status = status;
}

BogusChannel.prototype.suspend =
BogusChannel.prototype.resume =
function bc_suspres()
{
debug('bc_suspres\n');
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
}

var MozphoneModule = new Object();

MozphoneModule.registerSelf =
function mp_mod_registerSelf(compMgr, fileSpec, location, type)
{
debug('mp_mod_registerSelf\n');
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    var catman = Components.classes["@mozilla.org/categorymanager;1"]
        .getService(nsICategoryManager);
    
    debug("*** Registering -phone handler.\n");
    compMgr.registerFactoryLocation(CLINE_SERVICE_CID,
                                    "Mozphone CommandLine Service",
                                    CLINE_SERVICE_CONTRACTID, 
                                    fileSpec, location, type);
    catman.addCategoryEntry("command-line-argument-handlers",
                            "mozphone command line handler",
                            CLINE_SERVICE_CONTRACTID, true, true);
    
    debug("*** Registering content listener.\n");
    compMgr.registerFactoryLocation(TELCONTENT_LISTENER_CID,
                                    "TEL content listener",
                                    TELCONTENT_LISTENER_CONTRACTID, 
                                    fileSpec, location, type);
    catman.addCategoryEntry("external-uricontentlisteners",
                            TEL_MIMETYPE,
                            TELCONTENT_LISTENER_CONTRACTID, true, true);
    debug("*** Registering tel protocol handler.\n");
    compMgr.registerFactoryLocation(TELPROT_HANDLER_CID,
                                    "TEL protocol handler",
                                    TELPROT_HANDLER_CONTRACTID, 
                                    fileSpec, location, type);
    debug("*** Registering done.\n");
}

MozphoneModule.unregisterSelf =
function mp_mod_unregisterSelf(compMgr, fileSpec, location)
{
debug('mp_mod_unregisterSelf\n');
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    
    var catman = Components.classes["@mozilla.org/categorymanager;1"]
        .getService(nsICategoryManager);
    catman.deleteCategoryEntry("command-line-argument-handlers",
                               CLINE_SERVICE_CONTRACTID, true);
}

MozphoneModule.getClassObject =
function mp_mod_getClassObject(compMgr, cid, iid)
{
debug('mp_mod_getClassObject: ' + cid + ', iid=' + iid + '\n');
    // Checking if we're disabled in the Chrome Registry.
    var rv;
    try {
        var rdfSvc = Components.classes[RDFS_CONTRACTID].getService(nsIRDFService);
        var rdfDS = rdfSvc.GetDataSource("rdf:chrome");
        var resSelf = rdfSvc.GetResource("urn:mozilla:package:mozphone");
        var resDisabled = rdfSvc.GetResource("http://www.mozilla.org/rdf/chrome#disabled");
        rv = rdfDS.GetTarget(resSelf, resDisabled, true);
    } catch (e) {
debug( 'ERROR: ' + e + '\n' );
    }
if (rv)
  debug( 'Components.results.NS_ERROR_NO_INTERFACE');
    if (rv)
        throw Components.results.NS_ERROR_NO_INTERFACE;

    if (cid.equals(CLINE_SERVICE_CID))
        return CLineFactory;

    if (cid.equals(TELCONTENT_LISTENER_CID))
        return TELContentListenerFactory;
    if (cid.equals(TELPROT_HANDLER_CID))
        return TELProtocolHandlerFactory;
    if (!iid.equals(Components.interfaces.nsIFactory))
        throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    
    throw Components.results.NS_ERROR_NO_INTERFACE;
}

MozphoneModule.canUnload =
function mp_mod_canUnload(compMgr)
{
debug('mp_mod_canUnload\n');
    return true;
}

/* entrypoint */
function NSGetModule(compMgr, fileSpec)
{
debug('EntryPoint\n');
    return MozphoneModule;
}

